//
//  PutPinfInItsPlace.java
//  PutPinfInItsPlace
//
//  Created by Joshua King on 9/06/07.
//  Copyright (c) 2007 __MyCompanyName__. All rights reserved.
//
import java.io.*;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.*;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

public class PutPinfInItsPlace {

	private static final String[] ATOM_PATH = new String[] {
		"moov",
		"trak",
		"mdia",
		"minf",
		"stbl",
		"stsd",
		"mp4a",
		"esds"
	};
	
	private static final int MAGIC_NUMBER_BEFORE = 51;
	private static final int MAGIC_NUMBER_AFTER = 32819; 

	private static String insertBeforeExtension(String name, String insert) {
		if (name.lastIndexOf(".") != -1) {
			return name.substring(0, name.lastIndexOf(".")) + "." + insert + name.substring(name.lastIndexOf("."));
		} else {
			return name + "." + insert;
		}
	}

    public static void main (String args[]) {
		if (args.length == 0) {
			JOptionPane.showMessageDialog(null, "This tool fixes iTunes Plus AAC songs to play on Nokia mobile phones (and other devices).\n\nPlease select your songs in the following window.\nA backup will be made of each song, and then the song will be fixed.", "Fix Songs", JOptionPane.INFORMATION_MESSAGE);
			JFileChooser openDialog = new JFileChooser();
			FileFilter mp4aFilter = new FileFilter() {
				public boolean accept(File f) {
					return f.isDirectory() || (f.isFile() && f.getName().toLowerCase().endsWith(".m4a") && !f.getName().toLowerCase().endsWith(".purchased.m4a"));
				}
				
				public String getDescription() {
					return "iTunes Plus Unprotected Music (*.m4a)";
				}
			};
			openDialog.addChoosableFileFilter(mp4aFilter);
			openDialog.setMultiSelectionEnabled(true);
			openDialog.setDialogTitle("Choose Music to Fix");
			if (openDialog.showDialog(null, "Fix") != JFileChooser.APPROVE_OPTION) {
				return;
			}
			File[] files = openDialog.getSelectedFiles();
			args = new String[files.length];
			for (int i = 0; i < files.length; i++) {
				try {
					args[i] = files[i].getCanonicalPath();
				} catch (IOException ioe) {
					args[i] = null;
				}
			}
		}
		
		for (int i = 0; i < args.length; i++) {
			if (args[i] == null) continue;
			// First backup the file
			File source = new File(args[i]);
			File destination = new File(source.getParentFile(), insertBeforeExtension(source.getName(), "purchased"));
			if (!source.exists() || !source.isFile()) {
				JOptionPane.showMessageDialog(null, "Source file " + source + " does not appear to exist.\n\nCheck the directory exists, the file name is correct and that you have permission to read from the file.", "Fix Error", JOptionPane.ERROR_MESSAGE);
				continue;
			} else if (!source.canRead()) {
				JOptionPane.showMessageDialog(null, "Unable to read from source file " + source + ".\n\nCheck the directory exists, the file name is correct and that you have permission to read from the file.", "Fix Error", JOptionPane.ERROR_MESSAGE);
				continue;
			} else if (!source.canWrite()) {
				JOptionPane.showMessageDialog(null, "Unable to write to source file " + source + ".\n\nCheck the directory exists, the file name is correct and that you have permission to write to the file." , "Fix Error", JOptionPane.ERROR_MESSAGE);
				continue;
			} else if (destination.exists()) {
				JOptionPane.showMessageDialog(null, "Backup file " + destination + " already exists.\n\nIf you are sure you want to continue, please delete the backup file first.", "Fix Error", JOptionPane.ERROR_MESSAGE);
				continue;
			}
			// Try to move file
			else if (!source.renameTo(destination)) {
				JOptionPane.showMessageDialog(null, "Unable to backup " + source + " to " + destination + ".\n\nCheck you have permissions to write to the directory.", "Fix Error", JOptionPane.ERROR_MESSAGE);
				continue;
			}
			// Slowly try to work through all of the atoms
			FileChannel sourceChannel = null;
			FileChannel destinationChannel = null;
			try {
				sourceChannel = new FileInputStream(destination).getChannel();
				destinationChannel = new FileOutputStream(source).getChannel();
				int treepos = 0;
				boolean done = false;
				do {
					ByteBuffer atomSizeBuffer = ByteBuffer.allocate(4);
					ByteBuffer atomNameBuffer = ByteBuffer.allocate(4);
				
					// Get first atom
					if (sourceChannel.read(atomSizeBuffer) != 4 || sourceChannel.read(atomNameBuffer) != 4) {
						// File is short!
						throw new IOException("Short read from source file!");
					}
					atomSizeBuffer.flip();
					atomNameBuffer.flip();
					// Check name
					int atomSize = atomSizeBuffer.getInt();
					String atomName = new String(atomNameBuffer.array(), "ASCII");
					atomSizeBuffer.flip();
					atomNameBuffer.flip();
					if (done || !atomName.equals(ATOM_PATH[treepos])) {
						// Simply copy this atom as is
						if (sourceChannel.transferTo(sourceChannel.position() - 8, atomSize, destinationChannel) != atomSize) {
							// File is short!
							throw new IOException("Short read from source file!");
						}
						// Update position
						sourceChannel.position(sourceChannel.position() + atomSize - 8);
					} else {
						treepos++;
						if (treepos == ATOM_PATH.length) {
							// We've reached the desired atom
							if (atomSize == MAGIC_NUMBER_BEFORE) {
								atomSizeBuffer.putInt(atomSize = MAGIC_NUMBER_AFTER);
								atomSizeBuffer.flip(); // again
							} else {
								// Not fixing file!
							}
							done = true;
							// Copy (hopefully fixed) atom
							if (destinationChannel.write(atomSizeBuffer) != 4 || sourceChannel.transferTo(sourceChannel.position() - 4, atomSize - 4, destinationChannel) != atomSize - 4) {
								// File is short!
								throw new IOException("Short read from source file!");
							}
							// Update position
							sourceChannel.position(sourceChannel.position() + atomSize - 8);
						} else {
							// Copy head of atom
							if (sourceChannel.transferTo(sourceChannel.position() - 8, 8, destinationChannel) != 8) {
								// File is short!
								throw new IOException("Short read from source file!");
							}
							// Copy any extra bits needed (these aren't very validated)
							int extra = 0;
							if (atomName.equals("stsd")) extra = 8;
							else if (atomName.equals("mp4a")) extra = 28;
							if (extra > 0) {
								if (sourceChannel.transferTo(sourceChannel.position(), extra, destinationChannel) != extra) {
									// File is short!
									throw new IOException("Short read from source file!");
								}
								sourceChannel.position(sourceChannel.position() + extra);
							}							
						}
					}
				} while (sourceChannel.position() < sourceChannel.size());
				
				// Close files
				sourceChannel.close();
				destinationChannel.close();
				
			} catch (Exception e) {
				JOptionPane.showMessageDialog(null, "An exception was thrown while processing " + source + ".\n\n" + e.getMessage(), "Fix Error", JOptionPane.ERROR_MESSAGE);
				e.printStackTrace(System.err);
				// Revert files
				if (sourceChannel != null) try { sourceChannel.close(); } catch (Exception e2) {}
				if (destinationChannel != null) try { destinationChannel.close(); } catch (Exception e2) {}
				if (source.delete() && destination.renameTo(source)) {
					JOptionPane.showMessageDialog(null, "Unable to fix " + source + ".\n\nThe original file has been restored.", "Fix Error", JOptionPane.WARNING_MESSAGE);
				} else {
					JOptionPane.showMessageDialog(null, "Unable to fix " + source + ".\n\nThe original file could not be restored, but should exist as " + destination + ".\nRename this file to its original name, else iTunes will no longer play the song.", "Fix Error", JOptionPane.WARNING_MESSAGE);
				}
				
				continue;
			}
			
			JOptionPane.showMessageDialog(null, "Completed fix on " + source + ".\n\nThe original is stored as " + destination + ".", "Fix Complete", JOptionPane.INFORMATION_MESSAGE);
		}
    }
}
